// Copyright (c) 2017, Ravi Teja Gudapati. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:math';
import 'dart:async';
import 'package:jaguar/jaguar.dart';

/// Contains count information for a single throttle identifier/key
class ThrottleData {
  /// Count
  int count = 0;

  /// Start of the recorded time interval
  DateTime time;

  ThrottleData();

  ThrottleData.now(this.count) : time = new DateTime.now();

  ThrottleData.make(this.count, this.time);

  /// Increments the count
  ThrottleData inc(DateTime time) => new ThrottleData.make(count++, time);
}

/// Defines interface for throttle store
abstract class ThrottleStore {
  /// Registers a visit
  ///
  /// Increments the count. If the last
  FutureOr<ThrottleData> visit(String identifier, Duration interval);

  /// Fetches the throttle data for the given identifier
  FutureOr<ThrottleData> getData(String identifier);

  /// Resets the throttle data for the given identifier
  FutureOr reset(String identifier);
}

/// Throttle store backed by a Dart [Map]
///
/// Warning: Do not use it in distributed server
class MapThrottleStore implements ThrottleStore {
  //TODO: implement lock for accessing the store

  final Map<String, ThrottleData> _store = <String, ThrottleData>{};

  ThrottleData visit(String identifier, Duration interval) {
    final ThrottleData data = _store[identifier];

    if (data == null) {
      final ret = new ThrottleData.now(1);
      _store[identifier] = ret;
      return ret;
    }

    final now = new DateTime.now();

    if (now.difference(data.time) < interval) {
      _store[identifier] = data.inc(data.time);
      return data;
    }

    final ret = new ThrottleData.now(1);
    _store[identifier] = ret;
    return ret;
  }

  ThrottleData getData(String identifier) => _store[identifier];

  void reset(String identifier) {
    _store.remove(identifier);
  }
}

/// Identifies a HTTP request given the [Context] object
///
/// For example identifier could be based on
///   1. IP address
///   2. Logged-in User
///   3. Session id
///   4. route
///   5. Etc
///   6. Or any combination of the above
typedef FutureOr<String> ThrottleIdMaker(Context ctx);

/// [ThrottleIdMaker] to throttle requests by remote IP
String throttleIdByIp(Context ctx) =>
    ctx.req.connectionInfo.remoteAddress.address;

class Throttle extends Interceptor {
  /// Maximum number of requests allowed for the given [interval]
  final int quota;

  /// Interval
  final Duration interval;

  /// Identification maker. Creates an identification for the given request
  final ThrottleIdMaker idMaker;

  final ThrottleStore store;

  /// Provides an opportunity to create [Response] when rate-limit exceeds
  final RouteHandlerFunc failResponse;

  Throttle(this.quota, this.interval,
      {this.idMaker: throttleIdByIp, ThrottleStore store, this.failResponse})
      : store = store ?? new MapThrottleStore();

  _State _state;

  Future pre(Context ctx) async {
    final String throttleId = await idMaker(ctx);
    final ThrottleData count = await store.visit(throttleId, interval);

    _state = new _State.make(quota, count, interval);

    if (count == null) return;

    if (count.count > quota) {
      final Response resp = failResponse != null
          ? await failResponse(ctx)
          : new Response('Limit exceeded', statusCode: 429);

      throw _state.setHeaders(resp);
    }
  }

  Response post(Context ctx, Response incoming) => _state.setHeaders(incoming);
}

class _State {
  final int quota;

  final int remaining;

  final int resetsIn;

  final int retryAfter;

  _State(this.quota, this.remaining, this.resetsIn, this.retryAfter);

  factory _State.make(int quota, ThrottleData data, Duration interval) {
    if (data == null) {
      return new _State(quota, quota, interval.inSeconds, -1);
    }

    int remaining = max(0, quota - data.count);
    int resetsIn = max(0, data.time.difference(new DateTime.now()).inSeconds);
    int retryAfter = -1;
    if (quota <= data.count) {
      retryAfter = max(
          0, data.time.add(interval).difference(new DateTime.now()).inSeconds);
    }
    return new _State(quota, remaining, resetsIn, retryAfter);
  }

  Response setHeaders(Response resp) {
    resp.headers.set('X-RateLimit-Limit', quota.toString());
    resp.headers.set('X-RateLimit-Remaining', remaining.toString());
    resp.headers.set('X-RateLimit-Reset', resetsIn.toString());
    if (retryAfter >= 0) resp.headers.set('Retry-After', retryAfter.toString());
    return resp;
  }
}
